#Usage: python gic.py (your sequence file) (species: 'human' or 'mouse')
#The sequence length should be shorter than 20,000
#RNAfold is needed and should be added to your system path variable

#!/usr/bin/env python

import os
import sys
import subprocess
import re
import csv
import time
import math

csv.register_dialect('tab', delimiter='\t', quoting=csv.QUOTE_NONNUMERIC, lineterminator='\n')

LRMODEL_FEATURES_HUMAN = ['intercept', 'length', 'CGA', 'GCG', 'TCG', 'ACG', 'TCA', 'eng/len']
LRMODEL_COEFS_HUMAN = [0.7417, 2.612e-04, 48.66, 15.64, 76.23, -1.113, -60.29, 4.295] #Human, +/- sample ratio 1:1
LRMODEL_HUMAN = dict(zip(LRMODEL_FEATURES_HUMAN, LRMODEL_COEFS_HUMAN))

LRMODEL_FEATURES_MOUSE = ['intercept', 'length', 'CGA', 'GCG', 'TCG', 'ACG', 'TCA', 'eng/len']
LRMODEL_COEFS_MOUSE = [0.1625, 2.638e-04, 19.88, 37.59, 50.37, 35.44, -64.66, 2.194] #Mouse, +/- sample ratio 1:1
LRMODEL_MOUSE = dict(zip(LRMODEL_FEATURES_MOUSE, LRMODEL_COEFS_MOUSE))

LRMODEL_FEATURES = dict(zip(['human', 'mouse'], [LRMODEL_FEATURES_HUMAN, LRMODEL_FEATURES_MOUSE]))
LRMODELS = dict(zip(['human', 'mouse'], [LRMODEL_HUMAN, LRMODEL_MOUSE]))
BASES = ['A', 'T', 'C', 'G']
TRIPLETS = []
for B1 in BASES:
    for B2 in BASES:
        for B3 in BASES:
            TRIPLETS.append(B1+B2+B3)

def slidingWindow(seq, win, step=1):
    length = len(seq)
    mod = divmod((length-win), step)[1]
    if (win >= length):
        return seq
    else:
        start = 0
        end = win
        fragments = []
        while (len(seq[start:end]) == win):
            fragments.append(seq[start:end])
            start += step
            end += step
        if (mod > 0):
            fragments.append(seq[(length-win):])
        return fragments

def stat3mer(seq):
    tmp = []
    all3mer = slidingWindow(seq, win=3, step=1)
    for triplet in TRIPLETS:
        tmp.append(all3mer.count(triplet))
    num3mer = float(len(seq)-2)
    freq = dict(zip(TRIPLETS, [i/num3mer for i in tmp]))
    return freq

def calcSeqFeature(f):
    p = re.compile(r'(?P<eng>-?\d+\.\d+)')
    lens = []
    engs = []
    freqs = []
    with open(f, 'r') as fr:
        for row in fr:
            if row.startswith('>'):
                name = row[1:].strip().split()[0]
            elif ('(' in row):
                m = p.search(row)
                eng = float(m.groups()[0])
                engs.append(eng/lens[-1])
            else:
                lens.append(len(row.strip()))
                seq = row.strip().upper().replace('U', 'T')
                freq = stat3mer(seq)
                freqs.append([freq[_] for _ in LRMODEL_TRIPLETS])
    return lens, engs, freqs

def main(seqF):
    patn = re.compile(r'\.(txt|fst|fa)', flags=2)
    resF = patn.sub('_gic.txt', seqF)
    if (resF in os.listdir(os.getcwd())):
        print 'The results corresponding to %r have already been saved as the file %r.' % (seqF, resF)
    else:
        print 'STEP 2:\n\tRunning RNAfold... START @', time.ctime()
        ssF = patn.sub('_ss.txt', seqF)
        if (ssF not in os.listdir(os.getcwd())):
            cmd = 'RNAfold -noPS <%s >%s' % (seqF, ssF)
            sp = subprocess.call(cmd, shell=True)
        if (os.path.getsize(ssF) == 0):
            print 'STEP 2:\n\tPlease check whether RNAfold has been properly installed and configured. @', time.ctime()
            return
        print 'STEP 2:\n\tRunning RNAfold... DONE @', time.ctime()

        print 'STEP 3:\n\tCalculating sequence features... START @', time.ctime()
        seqLen, seqEng, seqLR3mer = calcSeqFeature(ssF)
        print 'STEP 3:\n\tCalculating sequence features... DONE @', time.ctime()

        print 'STEP 4:\n\tCalculating GIC score... START @', time.ctime()
        with open(resF, 'w') as fw:
            fields = ['name', 'GIC_score']+LRMODEL_FEATURE[1:]
            writer = csv.DictWriter(fw, fieldnames=fields, dialect='tab')
            writer.writerow(dict(zip(fields, fields)))
            with open(ssF, 'r') as fr:
                count = 0
                for row in fr:
                    if row.startswith('>'):
                        name = row[1:].strip().split()[0]
                        try:
                            features = [seqLen[count]]+seqLR3mer[count]+[seqEng[count]]
                        except IndexError, e:
                            break
                        features_dict = dict(zip(LRMODEL_FEATURE[1:], features))
                        tmp = LRMODEL['intercept']+sum([features_dict[_]*LRMODEL[_] for _ in LRMODEL_FEATURE[1:]])
                        gic = math.exp(tmp)/(math.exp(tmp)+1)
                        count += 1
                        writer.writerow(dict(zip(fields, [name, gic]+features)))
        print 'STEP 4:\n\tCalculating GIC score... DONE @', time.ctime()

if (__name__ == "__main__"):
    try:
        seqF = sys.argv[1]
        species = sys.argv[2]
        print 'STEP 1:\n\tLoading the %s RNA sequence file %r... DONE @' % (species, seqF), time.ctime()
        LRMODEL_FEATURE = LRMODEL_FEATURES[species]
        LRMODEL_TRIPLETS = LRMODEL_FEATURE[2:7]
        LRMODEL = LRMODELS[species]
        main(seqF)
    except IndexError, e:
        print '''Usage: python gic.py (your sequence file) (species: 'human' or 'mouse')
The sequence length should be shorter than 20,000'''